from datetime import datetime as Datetime
from decimal import Decimal

from openpyxl import Workbook


from chellow.gas.bill_parser_bgs_xlsx import _make_raw_bills, _parse_row
from chellow.utils import utc_datetime


def test_parse_row():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_values in enumerate(rows, start=1):
        for column, value in enumerate(row_values, start=1):
            ws.cell(row=row, column=column, value=value)

    row = 2
    title_row = tuple(ws.rows)[0]
    bills = {}
    _parse_row(bills, ws, row, title_row)


def test_make_raw_bills():
    wb = Workbook()
    ws = wb.active

    rows = [
        [
            "Customer",
            "Account",
            "MPRN",
            "Bill",
            "SupplyAddress",
            "BillDate",
            "BillingPeriod",
            "kWhs",
            "ChargeType",
            "ChargePeriodFrom",
            "ChargePeriodEnd",
            "Quantity",
            "QUnit",
            "Charge",
            "CUnit",
            "Total",
        ],
        [
            "Zorin Industries",
            69389283,
            86762618,
            584773269,
            "Extinct Volcano",
            Datetime(2025, 6, 13),
            Datetime(2025, 5, 1),
            0,
            "Gas",
            Datetime(2025, 5, 1),
            Datetime(2025, 5, 31),
            0,
            "kWh",
            0.5311,
            "p/kWh",
            0.00,
        ],
    ]

    for row, row_vals in enumerate(rows, start=1):
        for column, value in enumerate(row_vals, start=1):
            ws.cell(row=row, column=column, value=value)

    bills = _make_raw_bills(ws)
    expected_bills = [
        {
            "bill_type_code": "N",
            "mprn": "86762618",
            "reference": 584773269,
            "account": 69389283,
            "issue_date": utc_datetime(2025, 6, 12, 23, 0),
            "start_date": utc_datetime(2025, 4, 30, 23, 0),
            "finish_date": utc_datetime(2025, 5, 31, 22, 30),
            "kwh": Decimal("0"),
            "breakdown": {
                "admin_variable_gbp": Decimal("0.0"),
                "admin_variable_kwh": Decimal("0"),
                "admin_variable_rate": {Decimal("0.005311")},
            },
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": "['Zorin Industries', 69389283, 86762618, 584773269, 'Extinct "
            "Volcano', datetime.datetime(2025, 6, 13, 0, 0), "
            "datetime.datetime(2025, 5, 1, 0, 0), 0, 'Gas', "
            "datetime.datetime(2025, 5, 1, 0, 0), datetime.datetime(2025, "
            "5, 31, 0, 0), 0, 'kWh', 0.5311, 'p/kWh', 0.0]",
            "reads": [],
        }
    ]
    assert bills == expected_bills



from collections import defaultdict
from decimal import Decimal


from werkzeug.exceptions import BadRequest

from chellow.edi_lib import (
    ct_datetime,
    parse_edi,
    to_date,
    to_decimal,
    to_finish_date,
    to_utc,
)
from chellow.utils import HH, parse_mpan_core


read_type_map = {
    "00": "N",
    "09": "N3",
    "04": "C",
    "02": "E",
    "11": "E3",
    "01": "EM",
    "03": "W",
    "06": "X",
    "05": "CP",
    "12": "IF",
}


TMOD_MAP = {
    "139039": ("aahedc-gbp", "aahedc-rate", "aahedc-kwh"),
    "064305": ("fit-gbp", None, None),
    "590346": ("cfd-operational-gbp", None, None),
    "269100": ("bsuos-gbp", "bsuos-rate", "bsuos-kwh"),
    "422733": ("ccl-gbp", "ccl-rate", "ccl-kwh"),
    "273237": ("cfd-operational-gbp", "cfd-operational-rate", "cfd-operational-kwh"),
    "954379": ("cfd-interim-gbp", "cfd-interim-rate", "cfd-interim-kwh"),
    "538249": ("capaity-gbp", None, None),
    "568307": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "020330": ("eii-gbp", None, None),
    "439724": ("eii-gbp", None, None),
    "247610": ("eii-gbp", None, None),
    "930504": ("eii-gbp", None, None),
    "331201": ("eii-gbp", None, None),
    "307253": ("eii-gbp", "eii-rate", "eii-kwh"),
    "065950": ("eii-gbp", "eii-rate", "eii-kwh"),
    "095469": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "637050": ("meter-rental-gbp", "meter-rental-rate", "meter-rental-days"),
    "489920": ("elexon-gbp", "elexon-rate", "elexon-nbp-kwh"),
    "704107": ("fit-gbp", None, None),
    "019090": ("rego-gbp", "rego-rate", "rego-kwh"),
    "033667": ("management-gbp", "management-rate", "management-kwh"),
    "091890": ("shape-gbp", "shape-rate", "shape-kwh"),
    "122568": ("nrg-msp-gbp", "nrg-rate", "nrg-msp-kwh"),
    "716514": ("duos-amber-gbp", "duos-amber-rate", "duos-amber-kwh"),
    "769979": ("duos-red-gbp", "duos-red-rate", "duos-red-kwh"),
    "794486": ("capacity-gbp", "capacity-rate", "capacity-kwh"),
    "797790": ("duos-reactive-gbp", "duos-reactive-rate", "duos-reactive-kvarh"),
    "709522": (
        "duos-excess-availability-gbp",
        "duos-excess-availability-rate",
        "duos-excess-availability-kva",
    ),
    "644819": ("duos-fixed-gbp", "duos-fixed-rate", "duos-fixed-days"),
    "806318": ("duos-green-gbp", "duos-green-rate", "duos-green-kwh"),
    "209269": ("tnuos-gbp", "tnuos-rate", "tnuos-days"),
    "229128": ("ro-gbp", "ro-rate", "ro-kwh"),
    "012069": ("tnuos-gbp", None, None),
}

TPR_LOOKUP = {
    "Day": "00043",
    "Off Peak / Weekends": "00210",
    "Night": "00210",
    "Default Rate": "00043",
    "Single": "00210",
}


def _process_BCD(elements, headers):
    issue_date = to_date(elements["IVDT"][0])
    reference = elements["INVN"][0]
    bill_type_code = elements["BTCD"][0]

    headers["issue_date"] = issue_date
    headers["bill_type_code"] = bill_type_code
    headers["reference"] = reference


def _process_BTL(elements, headers):
    for bill in headers["bills"]:
        bill["mpan_core"] = headers["mpan_core"]
        bill["account"] = headers["account"]
        _customer_mods(headers, bill)
    return headers["bills"]


def _process_CCD1(elements, headers):
    tcod = elements["TCOD"]
    pres_read_date = to_finish_date(elements["PRDT"][0])

    prev_read_date = to_finish_date(elements["PVDT"][0])

    m = elements["MLOC"][0]
    mpan = " ".join((m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13]))

    prrd = elements["PRRD"]
    if len(prrd) < 4:
        return
    pres_read_type = read_type_map[prrd[1]]
    prev_read_type = read_type_map[prrd[3]]

    coefficient = Decimal(elements["ADJF"][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements["MTNR"][0]
    tpr_code = elements["TMOD"][0]
    if tpr_code == "kW":
        units = "kW"
        tpr_code = None
    elif tpr_code == "kVA":
        units = "kVA"
        tpr_code = None
    else:
        units = "kWh"
        tpr_code = TPR_LOOKUP[tcod[1]]

    try:
        reads = headers["reads"]
    except KeyError:
        reads = headers["reads"] = []

    reads.append(
        {
            "msn": msn,
            "mpan": mpan,
            "coefficient": coefficient,
            "units": units,
            "tpr_code": tpr_code,
            "prev_date": prev_read_date,
            "prev_value": prev_reading_value,
            "prev_type_code": prev_read_type,
            "pres_date": pres_read_date,
            "pres_value": pres_reading_value,
            "pres_type_code": pres_read_type,
        }
    )


def _process_CCD2(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TMOD"][0]
    headers["element_code"] = element_code
    try:
        eln_gbp, eln_rate, eln_cons = TMOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TMOD_MAP.")

    cons = elements["CONS"]
    kwh = Decimal("0")
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        if eln_gbp == "duos-availability-gbp":
            breakdown[eln_cons] = [el_cons]
        else:
            breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"]) / Decimal("100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"]) / Decimal("100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net
    breakdown["raw-lines"] = [headers["line"]]

    try:
        reads = headers["reads"]
        headers["reads"] = []
    except KeyError:
        reads = []

    bill = {
        "bill_type_code": headers["bill_type_code"],
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "start_date": start_date,
        "finish_date": finish_date,
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
    }
    headers["bills"].append(bill)


def _process_CCD3(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TMOD"][0]
    headers["element_code"] = element_code
    try:
        eln_gbp, eln_rate, eln_cons = TMOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TMOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons) / Decimal("1000")
        breakdown[eln_cons] = kwh = el_cons
    else:
        kwh = Decimal("0")

    bpri = elements["BPRI"]
    if len(bpri[0]) > 0:
        rate = to_decimal(bpri) / Decimal("100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"]) / Decimal("100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net
    breakdown["raw-lines"] = [headers["line"]]

    try:
        reads = headers["reads"]
        headers["reads"] = []
    except KeyError:
        reads = []

    bill = {
        "bill_type_code": headers["bill_type_code"],
        "issue_date": headers["issue_date"],
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "start_date": start_date,
        "finish_date": finish_date,
        "net": net,
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0"),
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
    }

    headers["bills"].append(bill)


def _process_CCD4(elements, headers):
    breakdown = defaultdict(int)

    element_code = elements["TMOD"][0]
    headers["element_code"] = element_code
    try:
        eln_gbp, eln_rate, eln_cons = TMOD_MAP[element_code]
    except KeyError:
        raise BadRequest(f"Can't find the element code {element_code} in the TMOD_MAP.")

    cons = elements["CONS"]
    if eln_cons is not None and len(cons[0]) > 0:
        el_cons = to_decimal(cons, "1000")
        breakdown[eln_cons] = kwh = el_cons

    if eln_rate is not None:
        rate = to_decimal(elements["BPRI"], "100000")
        breakdown[eln_rate] = [rate]

    start_date = to_date(elements["CSDT"][0])
    headers["bill_start_date"] = start_date

    finish_date = to_date(elements["CEDT"][0]) - HH
    headers["bill_finish_date"] = finish_date

    if "CTOT" in elements:
        net = Decimal("0.00") + to_decimal(elements["CTOT"], "100")
    else:
        net = Decimal("0.00")

    breakdown[eln_gbp] = net
    breakdown["raw-lines"] = [headers["line"]]

    try:
        reads = headers["reads"]
        del headers["reads"][:]
    except KeyError:
        reads = []

    bill = {
        "kwh": kwh if eln_gbp == "ro-gbp" else Decimal("0.00"),
        "reference": headers["reference"] + "_" + eln_gbp[:-4],
        "issue_date": headers["issue_date"],
        "start_date": start_date,
        "finish_date": finish_date,
        "net": net,
        "vat": Decimal("0.00"),
        "gross": net,
        "breakdown": breakdown,
        "reads": reads,
        "bill_type_code": headers["bill_type_code"],
    }
    headers["bills"].append(bill)


def _process_CDT(elements, headers):
    customer_id = elements["CIDN"][0]
    headers["customer_number"] = customer_id


def _process_CLO(elements, headers):
    cloc = elements["CLOC"]
    headers["account"] = cloc[1]


def _process_END(elements, headers):
    pass


def _process_MAN(elements, headers):
    madn = elements["MADN"]

    headers["mpan_core"] = parse_mpan_core("".join(madn[0:3]))


def _process_MHD(elements, headers):
    message_type = elements["TYPE"][0]
    if message_type == "UTLBIL":
        keep_keys = {"customer_number"}
        keep = {k: headers[k] for k in keep_keys}
        headers.clear()
        headers.update(keep)
        headers["bills"] = []


def _process_MTR(elements, headers):
    pass


def _process_VAT(elements, headers):
    vat = Decimal("0.00") + to_decimal(elements["UVTT"]) / Decimal("100")
    vat_percentage = to_decimal(elements["VATP"]) / Decimal("1000")
    vat_net = Decimal("0.00") + to_decimal(elements["UVLA"]) / Decimal("100")

    bill = {
        "bill_type_code": headers["bill_type_code"],
        "account": headers["account"],
        "mpan_core": headers["mpan_core"],
        "reference": headers["reference"] + "_vat",
        "issue_date": headers["issue_date"],
        "start_date": headers["bill_start_date"],
        "finish_date": headers["bill_finish_date"],
        "kwh": Decimal("0.00"),
        "net": Decimal("0.00"),
        "vat": vat,
        "gross": vat,
        "breakdown": {
            "raw-lines": [headers["line"]],
            "vat": {vat_percentage: {"vat": vat, "net": vat_net}},
        },
        "reads": [],
    }
    headers["bills"].append(bill)


def _process_NOOP(elements, headers):
    pass


CODE_FUNCS = {
    "BCD": _process_BCD,
    "BTL": _process_BTL,
    "CCD1": _process_CCD1,
    "CCD2": _process_CCD2,
    "CCD3": _process_CCD3,
    "CCD4": _process_CCD4,
    "CDT": _process_CDT,
    "CLO": _process_CLO,
    "DNA": _process_NOOP,
    "END": _process_END,
    "FIL": _process_NOOP,
    "MAN": _process_MAN,
    "MHD": _process_MHD,
    "MTR": _process_MTR,
    "SDT": _process_NOOP,
    "STX": _process_NOOP,
    "TYP": _process_NOOP,
    "TTL": _process_NOOP,
    "VAT": _process_VAT,
    "VTS": _process_NOOP,
}


def _customer_mods(headers, bill):
    if headers["customer_number"] == "WESSEXWAT":
        if (
            headers["element_code"] == "307660"
            and "ro-gbp" in bill["breakdown"]
            and bill["issue_date"] == to_utc(ct_datetime(2023, 4, 14))
            and bill["start_date"] == to_utc(ct_datetime(2023, 3, 1))
            and bill["finish_date"] == to_utc(ct_datetime(2023, 3, 31, 23, 30))
        ):
            bill["start_date"] = to_utc(ct_datetime(2021, 4, 1))
            bill["finish_date"] = to_utc(ct_datetime(2022, 3, 31, 23, 30))

    return bill


class Parser:
    def __init__(self, f):
        self.edi_str = str(f.read(), "utf-8", errors="ignore")
        self.line_number = None

    def make_raw_bills(self):
        bills = []
        headers = {"bills": []}
        for self.line_number, line, seg_name, elements in parse_edi(self.edi_str):
            headers["line"] = line
            try:
                func = CODE_FUNCS[seg_name]
            except KeyError:
                raise BadRequest(f"Code {seg_name} not recognized.")

            try:
                bills_chunk = func(elements, headers)
            except BadRequest as e:
                raise BadRequest(
                    f"{e.description} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                )
            except BaseException as e:
                raise BadRequest(
                    f"{e} on line {self.line_number} line {line} "
                    f"seg_name {seg_name} elements {elements}"
                ) from e

            if bills_chunk is not None:
                bills.extend(bills_chunk)

        return bills



import csv
from collections import defaultdict
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from enum import Enum, auto
from io import BytesIO


from openpyxl import load_workbook

from werkzeug.exceptions import BadRequest

from chellow.utils import c_months_u, to_ct, to_utc


class Title(Enum):
    CUSTOMER = auto()
    PRODUCT = auto()
    BROKER_NAME = auto()
    ACCOUNT = auto()
    MPRN = auto()
    BILL = auto()
    SUPPLY_ADDRESS = auto()
    BILL_DATE = auto()
    BILLING_PERIOD = auto()
    CHARGE_TYPE = auto()
    CHARGE_PERIOD_FROM = auto()
    CHARGE_PERIOD_END = auto()
    QUANTITY = auto()
    QUNIT = auto()
    CHARGE = auto()
    CUNIT = auto()
    TOTAL = auto()


COLUMNS = {
    Title.CUSTOMER: ["Customer"],
    Title.ACCOUNT: ["Account"],
    Title.MPRN: ["MPRN"],
    Title.BILL: ["Bill"],
    Title.BILL_DATE: ["Bill Date", "BillDate"],
    Title.BILLING_PERIOD: ["Billing Period", "BillingPeriod"],
    Title.CHARGE_TYPE: ["Charge Type", "ChargeType"],
    Title.CHARGE_PERIOD_FROM: ["Charge Period From", "ChargePeriodFrom"],
    Title.CHARGE_PERIOD_END: ["Charge Period End", "ChargePeriodEnd"],
    Title.QUANTITY: ["Quantity"],
    Title.QUNIT: ["QUnit"],
    Title.CHARGE: ["Charge"],
    Title.CUNIT: ["CUnit"],
    Title.TOTAL: ["Total"],
}


def make_column_map(title_row):
    titles = [cell.value for cell in title_row]
    column_map = {}
    for title, title_names in COLUMNS.items():
        idx = None
        for title_name in title_names:
            try:
                idx = titles.index(title_name)
            except ValueError:
                pass
        if idx is None:
            raise BadRequest(f"For the title {title} a column can't be found")

        column_map[title] = idx + 1
    return column_map


def get_date_naive(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None
    elif not isinstance(value, Datetime):
        raise BadRequest(
            f"Problem reading {value} as a timestamp at row {row} col {col}."
        )
    return value


def get_date(sheet, row, col_name):
    dt = get_date_naive(sheet, row, col_name)
    return None if dt is None else to_utc(to_ct(dt))


def get_value(sheet, row, col):
    try:
        return sheet.cell(row=row, column=col).value
    except IndexError:
        raise BadRequest(
            f"Can't find the cell at row {row} and col {col} on sheet {sheet}."
        )


def get_str(sheet, row, col):
    value = get_value(sheet, row, col)
    if value is None:
        return None

    return value.strip()


def get_dec(sheet, row, col):
    value = get_value(sheet, row, col)
    if value in ("", None):
        return None

    try:
        return Decimal(str(value))
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing the number '{value}' at row {row} col {col}. {e}"
        )


def get_int(sheet, row, col):
    value = get_value(sheet, row, col)
    try:
        return int(value)
    except ValueError as e:
        raise BadRequest(
            f"Problem parsing the integer '{value}' at row {row} col {col}. {e}"
        )


ELEMENT_LOOKUP = {
    "CCL": "ccl",
    "Fixed Management Fee": "admin_fixed",
    "Gas": "admin_variable",
    "Management Fee": "admin_variable",
    "Metering and Data": "metering",
    "LDZ Customer Capacity": "dn_customer_capacity_fixed",
    "LDZ Customer Fixed": "dn_customer_fixed",
    "LDZ System Capacity": "dn_system_capacity_fixed",
    "LDZ System Commodity": "dn_system_commodity",
    "NTS Exit Capacity (ECN)": "dn_ecn_fixed",
    "NTS SO Exit": "so_exit_commodity",
    "NTS TO Exit": "to_exit_commodity",
    "Unidentified Gas": "ug",
    "WAP": "commodity",
}

QUNIT_LOOKUP = {
    "kWh": "kwh",
    "days": "days",
}


def _parse_row(bills, sheet, row, title_row):
    column_map = make_column_map(title_row)
    mprn_raw = get_value(sheet, row, column_map[Title.MPRN])
    mprn = str(mprn_raw)
    reference = get_value(sheet, row, column_map[Title.BILL])
    account = get_value(sheet, row, column_map[Title.ACCOUNT])
    issue_date = get_date(sheet, row, column_map[Title.BILL_DATE])
    period_naive = get_date_naive(sheet, row, column_map[Title.BILLING_PERIOD])
    start_date, finish_date = list(
        c_months_u(start_year=period_naive.year, start_month=period_naive.month)
    )[0]

    try:
        mprn_values = bills[mprn]
    except KeyError:
        mprn_values = bills[mprn] = {}

    try:
        bill = mprn_values[period_naive]
    except KeyError:
        bill = mprn_values[period_naive] = {
            "bill_type_code": "N",
            "mprn": mprn,
            "reference": reference,
            "account": account,
            "issue_date": issue_date,
            "start_date": start_date,
            "finish_date": finish_date,
            "kwh": Decimal("0"),
            "breakdown": defaultdict(int, {}),
            "net_gbp": Decimal("0.00"),
            "vat_gbp": Decimal("0.00"),
            "gross_gbp": Decimal("0.00"),
            "raw_lines": str([(c.value) for c in sheet[row]]),
            "reads": [],
        }

    bd = bill["breakdown"]
    element_desc = get_value(sheet, row, column_map[Title.CHARGE_TYPE])
    quantity = get_dec(sheet, row, column_map[Title.QUANTITY])
    qunit = get_value(sheet, row, column_map[Title.QUNIT])
    charge = get_dec(sheet, row, column_map[Title.CHARGE]) / Decimal("100")
    total = get_dec(sheet, row, column_map[Title.TOTAL])
    if element_desc == "VAT":
        bill["net_gbp"] += quantity
        bill["vat_gbp"] += total
        bill["gross_gbp"] += quantity + total
        if "vat_rate" not in bd:
            bd["vat_rate"] = set()
        bd["vat_rate"].add(charge)
    else:
        element_name = ELEMENT_LOOKUP[element_desc]
        if element_name == "commodity":
            bill["kwh"] += quantity
        bd[f"{element_name}_gbp"] += total
        element_qunit = QUNIT_LOOKUP[qunit]
        bd[f"{element_name}_{element_qunit}"] += quantity
        rate_name = f"{element_name}_rate"
        if rate_name not in bd:
            bd[rate_name] = set()
        bd[rate_name].add(charge)


def _make_raw_bills(sheet):
    bills = {}
    rows = tuple(sheet.rows)
    title_row = rows[0]
    for row_index, row in enumerate(rows[1:], start=2):
        val = row[0].value
        if val not in (None, ""):
            try:
                _parse_row(bills, sheet, row_index, title_row)
            except BadRequest as e:
                raise BadRequest(f"On row {row_index}: {e.description}")

    raw_bills = []
    for mprn, mprn_values in bills.items():
        for period, raw_bill in mprn_values.items():
            raw_bills.append(raw_bill)

    return raw_bills


class Parser:
    def __init__(self, f):
        self.book = load_workbook(BytesIO(f), data_only=True)
        self.sheet = self.book.worksheets[0]

        self.last_line = None
        lines = (self._set_last_line(i, l) for i, l in enumerate(f))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        return _make_raw_bills(self.sheet)
